Require Import Comm.
Require Import Exec.
Require Import Relations.
Require Import Partial_Order.
Require Import relationOperators.
Require Import Classical.

Module ProgramFacts(P:PROGRAM).

  Import P E.

  Module ExecFacts := ExecutionFacts E.
  Export ExecFacts.

  Section jmm_proofs.

    Variable E:Exec.
  
    Theorem sw_po_consisntency : 
      rel_consistency (sw E) (po E).
    Proof.
      unfold rel_consistency.
      intros x y sw_x_y po_y_x.
      elim sw_x_y.
      intros so_x_y sameVar_x_y.
      exact (so_po_consistency E x y so_x_y po_y_x).
    Qed.

    Implicit Arguments trans_clos [A].

    Definition so_plus_po := fun x y:Act => so E x y \/ po E x y.
    
    Definition HB : relation Act := trans_clos so_plus_po.

    Lemma so_po_so : forall x y z w:Act, 
      so E x y -> po E y z -> so E z w -> 
      so E x w.
    Proof.
      intros x y z w so_x_y po_y_z so_z_w.
      

      destruct (so_synchronization E x y so_x_y) as (_ , y_synch).
      destruct (so_synchronization E z w so_z_w) as (z_synch, _).
      destruct (so_partial_order E) as (_, ord_trans, _).
      destruct (synch_so E y z (conj y_synch z_synch)) as [so_y_z|so_z_y].
      apply (ord_trans x z w (ord_trans x y z so_x_y so_y_z) so_z_w).
      
      rewrite (so_po_consistency E z y so_z_y po_y_z) in so_z_w.
      apply (ord_trans x y w so_x_y so_z_w).
    Qed.

    Lemma thread_discrimination : 
      forall x y:Act, 
        HB x y ->
        (act_thID x <> act_thID y ->
          exists x':Act, 
            exists y':Act, 
	      po E x x' /\ so E x' y' /\ po E y' y) /\
        (act_thID x = act_thID y -> po E x y).
    Proof.
      intros x y HB_x_y.
      destruct (po_partial_order E) as (po_refl, po_trans, _).
      
      (* The Proof goes by induction in the construction of the transitive 
         * closure. *)
      induction HB_x_y as 
        [x y so_plus_po_x_y | x y z so_plus_po_x_y HB_y_z HipInd_y_z].
      
      (* Base Case *)
      split.
      
      (* thID_act x <> thID_act y -> exists ...*)
      
      intro x_thNEq_y.
      destruct so_plus_po_x_y as [so_x_y | po_x_y]; exists x; exists y.
      split; [exact (po_refl x)|split; 
        [auto|exact (po_refl y)]].
      
      split; [exact (po_refl x)| split;[idtac|auto]].
      destruct (po_same_thread E x y po_x_y); tauto.
      
      (* thID_act x = thID_act y *)
      
      intro x_thEq_y.
      elim so_plus_po_x_y; [intro so_x_y|auto].
      destruct(same_thread_po E x y x_thEq_y) as [po_x_y|po_y_x]; 
        [auto|idtac].
      assert (x_eq_y:=so_po_consistency E x y so_x_y po_y_x); 
        rewrite x_eq_y; auto.
      
      (* Inductive Case *)
      destruct HipInd_y_z as (HipInd_thDiff, HipInd_thEq).
      split.
      
      (* thID_act x <> thID_act z *)
      
      intro x_NthEq_z.
      destruct so_plus_po_x_y as [so_x_y|po_x_y].
      
      (* so x y *)
      destruct (classic(act_thID y = act_thID z)) as 
        [y_thEq_z|y_NthEq_z].
      
      (* thID_act y = thID_act z *)
      assert (po_y_z := HipInd_thEq y_thEq_z).
      exists x; exists y.
      split; [exact (po_refl x)| split;[auto|auto]].
      
      (* thID_act y <> thID_act z *)
      destruct (HipInd_thDiff y_NthEq_z) as (x', exists_y'); 
        destruct exists_y' as (y', trans_y_z).
      decompose [and] trans_y_z.
      assert (so_x_y' := so_po_so x y x' y' so_x_y H H1); 
        exists x; exists y'.
      split; [exact (po_refl x)|auto].
      
      (* po x y *)
      
      assert (y_NthEq_z : act_thID y <> act_thID z); 
        [rewrite <- (po_same_thread E x y po_x_y); auto|idtac].
      destruct (HipInd_thDiff y_NthEq_z) as (x', exists_y'); 
        destruct exists_y' as (y', trans_y_z).
      decompose [and] trans_y_z; 
        assert (po_x_x' := po_trans x y x' po_x_y H).
      exists x';exists y'; auto.
      
      (* thID_act x = thID_act z *)
      intro x_thEq_z.
      destruct so_plus_po_x_y as [so_x_y|po_x_y].
      
      (* so x y *)
      destruct (classic(act_thID y = act_thID z)) as 
        [y_thEq_z|y_NthEq_z].
      (* thID_act y = thID_act z *)
      assert (x_thEq_y : act_thID x = act_thID y); 
        [rewrite y_thEq_z; auto|idtac].
      destruct (same_thread_po E x y x_thEq_y) as [po_x_y|po_y_x].
      exact (po_trans x y z po_x_y (HipInd_thEq y_thEq_z)).
      
      rewrite (so_po_consistency E x y so_x_y po_y_x).
      exact (HipInd_thEq y_thEq_z).
      
      (* thID_act y <> thID_act z *)
      destruct (HipInd_thDiff y_NthEq_z) as (x', exists_y'); 
        destruct exists_y' as (y', trans_y_z).
      decompose [and] trans_y_z; 
        assert (so_x_y' := so_po_so x y x' y' so_x_y H H1).
      assert (y'_thEq_z := po_same_thread E y' z H2).
      assert (x_thEq_y' : act_thID x = act_thID y'); 
        [rewrite y'_thEq_z; auto|idtac].
      destruct (same_thread_po E x y' x_thEq_y') as 
        [po_x_y'|po_y'_x].
      exact (po_trans x y' z po_x_y' H2).
      
      assert (x_eq_y' := so_po_consistency E x y' so_x_y' po_y'_x);
        rewrite x_eq_y'; auto.
      
      (* po x y *)
      assert (y_thEq_z : act_thID y = act_thID z); 
        [rewrite <- (po_same_thread E x y po_x_y); auto|idtac].
      exact (po_trans x y z po_x_y (HipInd_thEq y_thEq_z)).
    Qed.
    
    Lemma diff_thread_so : 
      forall x y:Act, 
        act_thID x <> act_thID y ->
        HB x y ->
        exists x':Act, 
          exists y':Act, 
	    po E x x' /\ so E x' y' /\ po E y' y.
    Proof.
      intros x y x_NthEq_y HB_x_y.
      destruct (thread_discrimination x y) as (th_diff, _);
        [auto|apply (th_diff x_NthEq_y)].
    Qed.
    
    Lemma sameThread_po : 
      forall x y:Act, 
        act_thID x = act_thID y ->
        HB x y -> po E x y.
    Proof.
      intros x y x_thEq_y HB_x_y.
      destruct (thread_discrimination x y) as (_, th_eq);
        [auto|apply (th_eq x_thEq_y)].
    Qed.

    Ltac cleanUp := repeat match goal with
                             | [H: so _ ?v ?v |- _ ] => clear H
                             | [H: po _ ?v ?v |- _ ] => clear H
                           end.

    Ltac ord_trans x y z ord ord_partial ord_x_z:= 
      match goal with
        | [h: ord _ x y, h1: ord _ y z |- _ ] => 
          destruct (ord_partial E) as (_, ord_tran, _); 
            assert(ord_x_z := ord_tran _ _ _ h h1); 
              clear ord_tran
      end.
    
    Ltac ord_antisym x y ord ord_partial eq_x_y := 
      match goal with
        | [h: ord _ x y, h1: ord _ y x |- _ ] => 
          destruct (ord_partial E) as (_ , _, ord_anti); 
            assert (eq_x_y := ord_anti _ _ h h1); 
              clear ord_anti
      end.

    Theorem HB_antisymmetric : antisymmetric Act HB.
    Proof.
      unfold antisymmetric; intros x y HB_x_y HB_y_x.
      destruct (Thread_ID.t_dec (act_thID x) (act_thID y)) as [x_thEq_y|x_NthEq_y].
      assert (po_x_y := sameThread_po _ _ x_thEq_y HB_x_y);
        assert (po_y_x := sameThread_po _ _ (sym_eq x_thEq_y) HB_y_x).
      ord_antisym x y po po_partial_order x_eq_y; auto.
      destruct (diff_thread_so x y x_NthEq_y HB_x_y) as (x', exists_y'); 
        destruct exists_y' as (y', trans_x_y);
          decompose [and] trans_x_y; clear trans_x_y.
      destruct (diff_thread_so y x (sym_not_eq x_NthEq_y) HB_y_x) as 
        (x'', exists_y''); 
        destruct exists_y'' as (y'', trans_y_x);
          decompose [and] trans_y_x; clear trans_y_x.
      ord_trans y' y x'' po po_partial_order po_y'_x''.
      assert(so_x'_y'' := so_po_so _ _ _ _ H1 po_y'_x'' H4).
      ord_trans y'' x x' po po_partial_order po_y''_x'.
      assert(so_x''_y' := so_po_so _ _ _ _ H4 po_y''_x' H1).
      assert (x'_eq_y'' := so_po_consistency _ _ _ so_x'_y'' po_y''_x'); 
        subst; cleanUp.
      assert (x''_eq_y' := so_po_consistency _ _ _ so_x''_y' po_y'_x''); 
        subst; cleanUp.
      ord_antisym y y' po po_partial_order y_eq_y'; subst; cleanUp.
      ord_antisym y'' x po po_partial_order y''_eq_x; subst; cleanUp.
      ord_antisym y' x so so_partial_order x_eq_y'; auto.
    Qed.


    Lemma sw_po_included_so_po : 
      relation_inclusion (fun x y:Act => sw E x y \/ po E x y) so_plus_po.
    Proof.
      unfold relation_inclusion; intros x y sw_po_x_y.
      unfold so_plus_po.
      destruct sw_po_x_y as [sw_x_y|po_x_y]; [unfold sw in sw_x_y; tauto|tauto].
    Qed.

    Lemma hb_included_HB : relation_inclusion  (hb E) HB.
    Proof.
      unfold hb,HB.
      intros x y hb_x_y.
      apply (clos_trans_trans_clos Act so_plus_po x y).
      exact ((rel_inclusion_trans_inclusion _ _ _) 
        sw_po_included_so_po _ _ hb_x_y).
    Qed.

    Lemma hb_antisymmetric : antisymmetric Act (hb E).
    Proof rel_inc_antisymm _ _ _ hb_included_HB HB_antisymmetric.

    Lemma hb_transitive : transitive Act (hb E).
    Proof.
      unfold hb, transitive.
      intros x y z clos_trans_x_y clos_trans_y_z.
      constructor 2 with y; [auto|auto].
    Qed.
    
    Lemma hb_reflexive : reflexive Act (hb E).
    Proof.
      unfold hb, reflexive.
      intros x.
      constructor 1.
      right. 
      destruct (po_partial_order E) as (po_refl, _, _); apply (po_refl x).
    Qed.

    Theorem hb_partial_order : Order (hb E).
    Proof. 
      constructor; 
        [exact hb_reflexive| 
          exact hb_transitive| 
            exact hb_antisymmetric].
    Qed.

  End jmm_proofs.
  
  Module Type TOP_SORT.
    
    Parameter T: Type.

    Definition in_rel (r:Relation T)(x:T) := 
      exists y:T, r x y \/ r y x.

    (* Axiomatization of the Topological Sort *)
    Parameter top_sort : forall r:Relation T, Order r -> Relation T.

    Section top_sort_axioms.
      
      Variable r:Relation T.
      Variable ord_r:Order r.

      Axiom top_sort_order : Order (top_sort r ord_r).

      Axiom ts_only_on_r :
        forall x y:T, top_sort r ord_r x y ->
          in_rel r x /\ in_rel r y.

      Axiom top_sort_total : 
        forall x y:T, in_rel r x -> in_rel r y ->
          top_sort r ord_r x y \/ top_sort r ord_r y x.

      Axiom top_sort_consistent : 
        rel_consistency r (top_sort r ord_r).

    End top_sort_axioms.
  End TOP_SORT.
  Declare Module top_sort : TOP_SORT with Definition T:=Act.
  Export top_sort.
End ProgramFacts.
