Require Import Relations.
Require Import Classical.

(*Just auxiliar definitions*)
(*Definition only_on_actions (E:Execution)(r:relation Actions) :=
  forall x y:Actions, r x y -> actions E x /\ actions E y.*)

Definition total_upto (A:Type)(r:relation A) (p: A -> Prop) : Prop :=
  forall x y:A, p x -> p y -> r x y \/ r y x.
Implicit Arguments total_upto [A].

Definition relation_inclusion (A:Type)(rel rel' :relation A) : Prop :=
  forall x y:A, rel x y -> rel' x y.

Implicit Arguments relation_inclusion [A].

Definition rel_consistency (A:Type)(r r':relation A) : Prop :=
  forall x y: A, r x y -> r' y x -> x = y.

Implicit Arguments rel_consistency [A].

Definition rel_consistency_restatement (A:Type)(r r':relation A) : Prop :=
  forall x y: A, x <> y -> r x y -> ~ r' y x.

Implicit Arguments rel_consistency_restatement [A].

Lemma counter_reciprocity : forall P Q:Prop, (P -> Q) -> (~Q -> ~P).
Proof.
  intros P Q P_then_Q not_Q.
  red.
  red in not_Q.
  intro P'.
  apply not_Q.
  apply P_then_Q.
  assumption.
Qed.

Lemma rel_cons_equiv : 
  forall (A:Type)(rel rel':relation A), 
    rel_consistency rel rel' <-> rel_consistency_restatement rel rel'.
Proof.
  intros A rel rel'.
  unfold rel_consistency_restatement, rel_consistency.
  split.
  
  intro rel_cons_res.
  intros x y x_neq_y rel_x_y.
  red; intro rel'_y_x.
  generalize (rel_cons_res x y rel_x_y rel'_y_x); intro x_eq_y.
  contradiction.
  
  intros rel_cons x y rel_x_y rel'_y_x.
  generalize (rel_cons x y).
  intro rel_cons_x_y.
  assert (h:=classic(x=y)).
  elim h; [auto|idtac].
  intro x_neq_y; generalize (rel_cons_x_y x_neq_y rel_x_y);
    intros not_rel'_y_x; contradiction.
Qed.

(*END Just auxiliar definitions*)


Section trans_clos.
  
  Variable A: Set.
  
  Inductive trans_clos (r:relation A) : relation A :=
    | base_trans : forall x y:A, r x y -> trans_clos r x y
    | ind_trans : forall x y z:A, r x y -> trans_clos r y z -> 
      trans_clos r x z.

  Implicit Arguments clos_trans [A].
  
  Lemma trans_clos_clos_trans : forall (r:relation A)(x y:A), 
    trans_clos r x y -> clos_trans r x y.
  Proof.
    intros r x y tc_x_y.
    induction tc_x_y.
    constructor 1; auto.
    
    assert (ct_x_y : clos_trans r x y); [constructor 1; auto|idtac].
    constructor 2 with y; [auto|auto].
  Qed.
  
  Lemma trans_clos_trans : forall (r:relation A)(x y z:A), 
    trans_clos r x y ->
    trans_clos r y z -> 
    trans_clos r x z.
  Proof.
    intros r x y z tc_x_y tc_y_z.
    induction tc_x_y.
    constructor 2 with y; [auto|auto].
    apply (ind_trans _ _ _ _ H (IHtc_x_y tc_y_z)).
  Qed.
  
  Lemma clos_trans_trans_clos : forall (r:relation A)(x y:A), 
    clos_trans r x y -> trans_clos r x y.
  Proof.
    intros r x y ct_x_y.
    induction ct_x_y; [constructor 1;auto|idtac].
    apply (trans_clos_trans _ _ _ _ IHct_x_y1 IHct_x_y2).
  Qed.
  
  Lemma trans_equiv : forall (r:relation A)(x y:A), 
    clos_trans r x y <-> trans_clos r x y.
  Proof.
    intros r x y.
    split; [apply (clos_trans_trans_clos r x y)| 
      apply (trans_clos_clos_trans r x y)].
  Qed.
  
  Lemma rel_inclusion_trans_inclusion : 
    forall (r r1:relation A), 
      relation_inclusion r r1 -> 
      relation_inclusion (clos_trans r)(clos_trans r1).
  Proof.
    intros r r1; unfold relation_inclusion.
    intros rel_inc_r_r1 x y trans_r_x_y.
    induction trans_r_x_y as 
      [x y base|x y z trans_x_y IH_x_y trans_y_z IH_y_z].
    
    constructor 1; apply (rel_inc_r_r1 _ _  base).
    
    constructor 2 with y; [auto|auto].
  Qed.
  
  Lemma rel_inc_antisymm : 
    forall r r1:relation A, 
      relation_inclusion r r1 -> 
      antisymmetric A r1 -> 
      antisymmetric A r.
  Proof.
    unfold antisymmetric.
    intros r r1 inc_r_r1 antisymm_r1.
    intros x y r_x_y r_y_x.
    apply (antisymm_r1 _ _ (inc_r_r1 _ _ r_x_y)(inc_r_r1 _ _ r_y_x)).
  Qed.

End trans_clos.
